#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <unistd.h>
#include <sys/wait.h>
#include <semaphore.h>

#define SHM_SIZE 1024  /* 1KB delt hukommelse segment */

/*
Delte datastruktur. Indeholer en POSIX semafor og en delt
streng.
*/
typedef struct {
    sem_t sem;
    char str[];
} shared_t;

int main(int argc, char *argv[])
{
    pid_t pid;
    int shmid;
    shared_t *s;

    /* Opret et delt hukommelsessegment med strrelsen SHM_SIZE */
    if ((shmid = shmget(IPC_PRIVATE, sizeof(shared_t) + SHM_SIZE, 0600 | IPC_CREAT)) == -1) {
        perror("shmget");
        exit(1);
    }

    /* Lg segmentet ind i processens adresserum. 'data' peger nu p
    den delte hukommelse */
    s = shmat(shmid, NULL, 0);
    if (s == (void *)(-1)) {
        perror("shmat");
        exit(1);
    }

    /* Initialiser semafor til 0. Dvs. sem_wait() blokerer indtil
       sem_post() bliver kaldt */
    if (sem_init(&s->sem, 1, 0) == -1) {
        perror("sem_init");
        exit(1);
    }

    /* Del i to processer */
    pid = fork();

    /* Child proces */
    if (pid == 0) {
        sem_wait(&s->sem);
        printf("segment indeholder: \"%s\"\n", s->str);
    }
    /* Parent proces */
    else {
        strncpy(s->str,
            "Pas p med at tilg delt hukommelse uden synkronisering!!!",
        SHM_SIZE);
        sem_post(&s->sem);

        wait(NULL); // Vent p at child terminerer.
    }

    /* Udtag segmentet fra adresserummet */
    if (shmdt(s) == -1) {
        perror("shmdt");
        exit(1);
    }

    return 0;
}
